/*
htc.c

Copyright (C) 1999 Lars Brinkhoff.  See COPYING for terms and conditions.

htc is the client half of httptunnel.  httptunnel creates a virtual
two-way data path tunneled in HTTP requests.
*/

#include <stdio.h>
#include <fcntl.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <sys/poll.h>
#include <sys/socket.h>

#include "common.h"

typedef struct
{
  char *me;
  char *device;
  char *host_name;
  int host_port;
  char *proxy_name;
  int proxy_port;
  size_t proxy_buffer_size;
  int proxy_buffer_timeout;
  size_t content_length;
  int forward_port;
} Arguments;

#define NO_PROXY_BUFFER 0
#define NO_PROXY (NULL)

static void
usage (FILE *f, const char *me)
{
  fprintf (f,
"Usage: %s [OPTION]... HOSTNAME:PORT\n"
"  -B, --proxy-buffer-size BYTES  ...\n"
"  -c, --content-length BYTES     use HTTP PUT requests of BYTES size\n"
"  -d, --device DEVICE            use DEVICE for input and output\n"
"  -f, --forward-port             use TCP port PORT for input and output\n"
"  -h, --help                     display this help and exit\n"
"  -P, --proxy HOSTNAME:PORT      use a HTTP proxy\n"
"  -T, --timeout TIME             timeout, in milliseconds, before sending\n"
"                                 padding to a buffering proxy\n"
"  -V, --version                  output version information and exit\n",
	   me);
}

static int
wait_for_connection_on_socket (int s)
{
  struct sockaddr addr;
  socklen_t len;
  int t;

  len = sizeof addr;
  t = accept (s, &addr, &len);
  if (t == -1)
    return -1;

  return t;
}

static int
atoi_with_postfix (const char *s_)
{
  char *s = strdup (s_);
  int n = strlen (s);
  int factor = 1;
  int x;

  if (s == NULL)
    {
      fprintf (stderr, "Out of memory\n");
      exit (1);
    }

  switch (s[n - 1])
    {
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
      break;
    case 'k':
    case 'K':
      factor = 1024;
      break;
    case 'M':
      factor = 1024 * 1024;
      break;
    case 'G':
      factor = 1024 * 1024 * 1024;
      break;
    default:
      fprintf (stderr, "Unknown prefix: %c\n", s[n - 1]);
      exit (1);
    }

  if (factor != 1)
    s[n - 1] = '\0';

  x = factor * atoi (s);
  free (s);
  return x;
}

static void
parse_arguments (int argc, char **argv, Arguments *arg)
{
  int c;

  /* defaults */

  arg->me = argv[0];
  arg->device = NULL;
  arg->forward_port = -1;
  arg->host_name = NULL;
  arg->host_port = -1;
  arg->proxy_name = NO_PROXY;
  arg->proxy_port = -1;
  arg->proxy_buffer_size = NO_PROXY_BUFFER;
  arg->proxy_buffer_timeout = 200;
  arg->content_length = 1000000;
  
  for (;;)
    {
      int option_index = 0;
      static struct option long_options[] =
      {
	{ "help", no_argument, 0, 'h' },
	{ "version", no_argument, 0, 'V' },
	{ "proxy", required_argument, 0, 'P' },
	{ "device", required_argument, 0, 'd' },
	{ "timeout", required_argument, 0, 'T' },
	{ "forward-port", required_argument, 0, 'f' },
	{ "content-length", required_argument, 0, 'c' },
	{ "proxy-buffer-size", required_argument, 0, 'B' },
	{ 0, 0, 0, 0 }
      };

      c = getopt_long (argc, argv, "B:c:d:f:hP:T:V",
		       long_options, &option_index);
      if (c == -1)
	break;

      switch (c)
	{
	case 0:
	  printf ("option %s", long_options[option_index].name);
	  if (optarg)
	    printf (" with arg %s", optarg);
	  printf ("\n");
	  break;

	case 'B':
	  arg->proxy_buffer_size = atoi_with_postfix (optarg);
	  break;

	case 'c':
	  arg->content_length = atoi_with_postfix (optarg);
	  break;

	case 'd':
	  arg->device = optarg;
	  break;

	case 'f':
	  arg->forward_port = atoi (optarg);
	  break;

	case 'h':
	  usage (stdout, arg->me);
	  exit (0);

	case 'P':
	  name_and_port (optarg, &arg->proxy_name, &arg->proxy_port);
	  break;

	case 'T':
	  arg->proxy_buffer_timeout = atoi (optarg);
	  break;

	case 'V':
	  printf ("htc (%s) %s\n", PACKAGE, VERSION);
	  exit (0);

	case '?':
	  break;

	default:
	  printf ("?? getopt returned character code 0%o ??\n", c);
	}
    }

  if (optind == argc - 1)
    name_and_port (argv[optind], &arg->host_name, &arg->host_port);
  else
    {
      fprintf (stderr, "%s: the destination of the tunnel must be specified.\n"
	               "%s: try '%s --help' for help.\n",
	       arg->me, arg->me, arg->me);
      exit (1);
    }

  if (arg->device == NULL && arg->forward_port == -1)
    {
      fprintf (stderr, "%s: one of --device or --forward-port must be used.\n"
	               "%s: try '%s -help' for help.\n",
	       arg->me, arg->me, arg->me);
      exit (1);
    }

  if (arg->device != NULL && arg->forward_port != -1)
    {
      fprintf (stderr, "%s: --device can't be used together with "
	                   "--forward-port.\n"
	               "%s: try '%s --help' for help.\n",
	       arg->me, arg->me, arg->me);
    }

  if (((arg->device == NULL) == (arg->forward_port == -1)) ||
      arg->host_name == NULL ||
      arg->host_port == -1 ||
      ((arg->proxy_name == NO_PROXY) != (arg->proxy_port == -1)))
    {
      usage (stderr, arg->me);
      exit (1);
    }

  if (arg->proxy_name == NO_PROXY)
    {
      if (arg->proxy_buffer_size != NO_PROXY_BUFFER)
	{
	  fprintf (stderr, "%s: warning: --proxy-buffer-size can't be "
		   "used without --proxy\n", arg->me);
	  arg->proxy_buffer_size = NO_PROXY_BUFFER;
	}

      if (arg->proxy_buffer_timeout != -1)
	{
	  fprintf (stderr, "%s: warning: --proxy-buffer-timeout can't be "
		   "used without --proxy\n", arg->me);
	  arg->proxy_buffer_timeout = -1;
	}
    }
  else if (arg->proxy_buffer_size == NO_PROXY_BUFFER)
    arg->proxy_buffer_timeout = -1;
}

int
main (int argc, char **argv)
{
  int s = -1;
  int fd = -1;
  Arguments arg;
  Tunnel *tunnel;
  int closed;

  parse_arguments (argc, argv, &arg);

  if (arg.forward_port != -1)
    {
      s = server_socket (arg.forward_port);
      if (s == -1)
	{
	  fprintf (stderr, "%s: couldn't create server socket: %s\n",
		   arg.me, strerror (errno));
	  exit (1);
	}
    }

  if (fork () != 0)
    exit (0);

  close (0);
  close (1);
  /* close (2); */

  for (;;)
    {
      if (arg.device)
	{
	  fd = open_device (arg.device);
	  if (fd == -1)
	    {
	      fprintf (stderr, "%s: couldn't open %s: %s\n",
		       arg.me, arg.device, strerror (errno));
	      exit (1);
	    }
	}
      else if (arg.forward_port != -1)
	{
	  fd = wait_for_connection_on_socket (s);
	  if (fd == -1)
	    {
	      fprintf (stderr, "%s: couldn't forward port %d: %s\n",
		       arg.me, arg.forward_port, strerror (errno));
	      exit (1);
	    }
	}

      tunnel = tunnel_new_client (arg.host_name, arg.host_port,
				  arg.proxy_name, arg.proxy_port,
				  arg.content_length);
      if (tunnel == NULL)
	{
	  fprintf (stderr, "%s: couldn't create tunnel\n", arg.me);
	  exit (1);
	}

      if (tunnel_connect (tunnel) == -1)
	{
	  fprintf (stderr, "couldn't open tunnel: %s\n", strerror (errno));
	  exit (1);
	}

      closed = FALSE;
      while (!closed)
	{
	  struct pollfd pollfd[2];
	  int n;

	  pollfd[0].fd = fd;
	  pollfd[0].events = POLLIN;
	  pollfd[1].fd = tunnel_in_fd (tunnel);
	  pollfd[1].events = POLLIN;
      
	  n = poll (pollfd, 2, arg.proxy_buffer_timeout);
	  if (n == -1)
	    {
	      perror ("poll");
	      exit (1);
	    }
	  else if (n == 0)		/* timeout occurred */
	    {
	      if (arg.proxy_buffer_size != NO_PROXY_BUFFER)
		tunnel_maybe_pad (tunnel, arg.proxy_buffer_size);
	      continue;
	    }
      
	  if (pollfd[0].revents)
	    {
	      ssize_t n;

	      n = handle_device_input (fd,
				       pollfd[0].revents,
				       tunnel);
	      if (n == 0 || (n == -1 && errno != EAGAIN))
		{
		  closed = TRUE;
		}		
	    }

	  if (pollfd[1].revents)
	    {
	      ssize_t n;

	      n = handle_tunnel_input (tunnel,
				       pollfd[1].revents,
				       fd);
	      if (n == 0 || (n == -1 && errno != EAGAIN))
		{
		  closed = TRUE;
		}		
	    }
	}

      close (fd);
      tunnel_close (tunnel);
    }

  close (s);
  tunnel_destroy (tunnel);

  exit (0);
}
