/*
hts.c

Copyright (C) 1999 Lars Brinkhoff.  See COPYING for terms and conditions.

hts is the server half of httptunnel.  httptunnel creates a virtual
two-way data path tunneled in HTTP requests.
*/

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/poll.h>

#include "common.h"

typedef struct
{
  char *me;
  char *device;
  int port;
  char *forward_host;
  int forward_port;
} Arguments;

static void
usage (FILE *f, const char *me)
{
  fprintf (f,
"Usage: %s [OPTION]... PORT\n"
"  -d, --device DEVICE            use DEVICE for input and output\n"
"  -f, --forward-port HOST:PORT   connect to PORT at HOST and use\n"
"                                 for input and output\n"
"  -h, --help                     display this help and exit\n"
"  -V, --version                  output version information and exit\n",
	   me);
}

static void
parse_arguments (int argc, char **argv, Arguments *arg)
{
  int c;

  /* defaults */

  arg->me = argv[0];
  arg->device = NULL;
  arg->forward_host = NULL;
  arg->forward_port = -1;
  
  for (;;)
    {
      int option_index = 0;
      static struct option long_options[] =
      {
	{ "help", no_argument, 0, 'h' },
	{ "version", no_argument, 0, 'V' },
	{ "device", required_argument, 0, 'd' },
	{ "forward-port", required_argument, 0, 'f' },
	{ 0, 0, 0, 0 }
      };

      c = getopt_long (argc, argv, "d:f:hV",
		       long_options, &option_index);
      if (c == -1)
	break;

      switch (c)
	{
	case 0:
	  printf ("option %s", long_options[option_index].name);
	  if (optarg)
	    printf (" with arg %s", optarg);
	  printf ("\n");
	  break;

	case 'd':
	  arg->device = optarg;
	  break;

	case 'f':
	  name_and_port (optarg, &arg->forward_host, &arg->forward_port);
	  break;

	case 'h':
	  usage (stdout, arg->me);
	  exit (0);

	case 'V':
	  printf ("hts (%s) %s\n", PACKAGE, VERSION);
	  exit (0);

	case '?':
	  break;

	default:
	  printf ("?? getopt returned character code 0%o ??\n", c);
	}
    }

  if (optind == argc - 1)
    arg->port = atoi (argv[optind]);
  else
    {
      fprintf (stderr, "%s: the port number used to a accept tunnel must "
	                   "be specified.\n"
	               "%s: try '%s --help' for help.\n",
	       arg->me, arg->me, arg->me);
      exit (1);
    }

  if (arg->device == NULL && arg->forward_port == -1)
    {
      fprintf (stderr, "%s: one of --device or --forward-port must be used.\n"
	               "%s: try '%s -help' for help.\n",
	       arg->me, arg->me, arg->me);
      exit (1);
    }

  if (arg->device != NULL && arg->forward_port != -1)
    {
      fprintf (stderr, "%s: --device can't be used together with "
	                   "--forward-port.\n"
	               "%s: try '%s --help' for help.\n",
	       arg->me, arg->me, arg->me);
    }

  if (((arg->device == NULL) == (arg->forward_port == -1)) ||
      arg->port == -1 ||
      ((arg->forward_host == NULL) != (arg->forward_port == -1)))
    {
      usage (stderr, arg->me);
      exit (1);
    }
}

int
main (int argc, char **argv)
{
  int closed;
  int fd = -1;
  Arguments arg;
  Tunnel *tunnel;

  parse_arguments (argc, argv, &arg);

  if (fork () != 0)
    exit (0);

  close (0);
  close (1);
  /* close (2); */

  tunnel = tunnel_new_server (arg.port);
  if (tunnel == NULL)
    {
      fprintf (stderr, "%s: couldn't create tunnel\n", argv[0]);
      exit (1);
    }

  for (;;)
    {
      if (arg.device != NULL)
	{
	  fd = open_device (arg.device);
	  if (fd == -1)
	    {
	      fprintf (stderr, "%s: couldn't open %s: %s\n",
		       arg.me, arg.device, strerror (errno));
	      exit (1);
	    }
	}

      if (tunnel_accept (tunnel) == -1)
	{
	  fprintf (stderr, "%s: couldn't establish connection: %s\n",
		   argv[0], strerror (errno));
	  exit (1);
	}

      if (arg.forward_port != -1)
	{
	  struct sockaddr_in addr;

	  if (set_address (&addr, arg.forward_host, arg.forward_port) == -1)
	    {
	      fprintf (stderr, "%s: couldn't forward port to %s:%d: %s\n",
		       arg.me, arg.forward_host, arg.forward_port,
		       strerror (errno));
	      exit (1);
	    }

	  fd = do_connect (&addr);
	  if (fd == -1)
	    {
	      fprintf (stderr, "%s: couldn't connect to %s:%d: %s\n",
		       arg.me, arg.forward_host, arg.forward_port,
		       strerror (errno));
	      exit (1);
	    }
	}

      closed = FALSE;
      while (!closed)
	{
	  struct pollfd pollfd[2];
	  int n;
      
	  pollfd[0].fd = fd;
	  pollfd[0].events = POLLIN;
	  pollfd[1].fd = tunnel_in_fd (tunnel);
	  pollfd[1].events = POLLIN;

	  n = poll (pollfd, 2, -1);
	  if (n == -1)
	    {
	      perror ("poll");
	      exit (1);
	    }
      
	  if (pollfd[0].revents)
	    {
	      ssize_t n;

	      n = handle_device_input (fd,
				       pollfd[0].revents,
				       tunnel);
	      if (n == 0 || (n == -1 && errno != EAGAIN))
		{
		  closed = TRUE;
		}		
	    }

	  if (pollfd[1].revents)
	    {
	      ssize_t n;

	      n = handle_tunnel_input (tunnel,
				       pollfd[1].revents,
				       fd);
	      if (n == 0 || (n == -1 && errno != EAGAIN))
		{
		  closed = TRUE;
		}		
	    }
	}

      fsync (fd);
      close (fd);
      tunnel_close (tunnel);
    }

  tunnel_destroy (tunnel);
 
  return 0;
}
