/*
tunnel.c

Copyright (C) 1999 Lars Brinkhoff.  See COPYING for terms and conditions.

Tunnel *tunnel_new_client (const char *host, int host_port,
		    const char *proxy, int proxy_port, size_t content_length);

  Create a new HTTP tunnel client.

Tunnel *tunnel_new_server (int port);

  Create a new HTTP tunnel server.

int tunnel_connect (Tunnel *tunnel);

  Open the tunnel.  (Client only.)

int tunnel_accept (Tunnel *tunnel);

  Accept a tunnel connection.  (Server only.)

int tunnel_in_fd (Tunnel *tunnel);
int tunnel_out_fd (Tunnel *tunnel);

  Return the file descriptor used for reading or writing, respectively.
  Warning: don't try to use these with read() or write().  Currently,
  they are only meant to be used with poll() or select().

ssize_t tunnel_read (Tunnel *tunnel, void *data, size_t length);
ssize_t tunnel_write (Tunnel *tunnel, void *data, size_t length);

  Read or write to the tunnel.  Same semantics as with read() and
  write().  Watch out for return values less than length.

int tunnel_padding (Tunnel *tunnel, size_t length);

  Send length pad bytes.

int tunnel_maybe_pad (Tunnel *tunnel, size_t length);

  Pad to nearest even multiple of length.

int tunnel_close (Tunnel *tunnel);

  Close the tunnel.
*/

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "tunnel.h"
#include "common.h"

#define TUNNEL_IN 1
#define TUNNEL_OUT 2

#define min(a, b) ((a) < (b) ? (a) : (b))

struct tunnel
{
  int in_fd, out_fd;
  int server_socket;
  const char *host, *proxy;
  int host_port, proxy_port;
  struct sockaddr_in address;
  size_t bytes;
  size_t content_length;
  char buf[65536];
  char *buf_ptr;
  size_t buf_len;
};

static const size_t sizeof_header = sizeof (Request) + sizeof (Length);

static inline int
tunnel_is_disconnected (Tunnel *tunnel)
{
  return tunnel->out_fd == -1;
}

static inline int
tunnel_is_connected (Tunnel *tunnel)
{
  return !tunnel_is_disconnected (tunnel);
}

static inline int
tunnel_is_server (Tunnel *tunnel)
{
  return tunnel->host == NULL;
}

static inline int
tunnel_is_client (Tunnel *tunnel)
{
  return !tunnel_is_server (tunnel);
}

static inline int
http_method (Tunnel *tunnel, int fd, const char *method, ssize_t length)
{
  char str[1000];

  if (fd == -1)
    return -1;

  if (tunnel->proxy == NULL)
    {
      if (length >= 0)
	{
	  sprintf (str,
		   "%s /index.html HTTP/1.0\r\n"
		   "Host: %s:%d\r\n"
		   "Content-Length: %d\r\n"
		   "\r\n",
		   method,
		   tunnel->host, tunnel->host_port,
		   length);
	}
      else
	{
	  sprintf (str,
		   "%s /index.html HTTP/1.0\r\n"
		   "Host: %s:%d\r\n"
		   "\r\n",
		   method,
		   tunnel->host, tunnel->host_port);
	}
    }
  else
    {
      if (length >= 0)
	{
	  sprintf (str,
		   "%s http://%s:%d/index.html HTTP/1.0\r\n"
		   "Host: %s:%d\r\n"
		   "Content-Length: %d\r\n"
		   "\r\n",
		   method,
		   tunnel->host, tunnel->host_port,
		   tunnel->host, tunnel->host_port,
		   length);
	}
      else
	{
	  sprintf (str,
		   "%s http://%s:%d/index.html HTTP/1.0\r\n"
		   "Host: %s:%d\r\n"
		   "\r\n",
		   method,
		   tunnel->host, tunnel->host_port,
		   tunnel->host, tunnel->host_port);
	}
    }

  if (write_all (fd, str, strlen (str)) == -1)
    return -1;

  return 0;
}

static int
http_get (Tunnel *tunnel, int fd)
{
  return http_method (tunnel, fd, "GET", (ssize_t)-1);
}

static int
http_put (Tunnel *tunnel, int fd, size_t length)
{
  return http_method (tunnel, fd, "PUT", (ssize_t)length);
}

#if 0
static int
http_post (int fd, size_t length)
{
  return http_method (fd, "POST", (ssize_t)length);
}
#endif

static void
tunnel_disconnect (Tunnel *tunnel)
{
  if (tunnel_is_disconnected (tunnel))
    return;

  if (tunnel->bytes != tunnel->content_length)
    {
      fprintf (stderr, "tunnel_disconnect: warning: "
	       "bytes=%d != content_length=%d\n",
	       tunnel->bytes, tunnel->content_length);
    }

  shutdown (tunnel->out_fd, 1);
  fsync (tunnel->out_fd);
  close (tunnel->out_fd);

  tunnel->out_fd = -1;
  tunnel->bytes = 0;
}

static int
tunnel_out_connect (Tunnel *tunnel)
{
  if (!tunnel_is_disconnected (tunnel))
    {
      fprintf (stderr, "tunnel_out_connect: was connected\n");
      tunnel_disconnect (tunnel);
    }

  tunnel->out_fd = do_connect (&tunnel->address);
  if (tunnel->out_fd == -1)
    return -1;

  shutdown (tunnel->out_fd, 0);

  http_put (tunnel, tunnel->out_fd, tunnel->content_length);

  tunnel->bytes = 0;

  return 0;
}

static int
tunnel_in_connect (Tunnel *tunnel)
{
  if (tunnel->in_fd != -1)
    return -1;

  tunnel->in_fd = do_connect (&tunnel->address);
  if (tunnel->in_fd == -1)
    return -1;

  http_get (tunnel, tunnel->in_fd);

  shutdown (tunnel->in_fd, 1);

  return 0;
}

static inline ssize_t
tunnel_write_data (Tunnel *tunnel, void *data, size_t length)
{
  if (write_all (tunnel->out_fd, data, length) == -1)
    return -1;
  tunnel->bytes += length;
  return length;
}

int
tunnel_write_request (Tunnel *tunnel, Request request,
		      void *data, Length length)
{
  if (tunnel->bytes + sizeof request +
      (length > 0 ? sizeof length + length : 0) > tunnel->content_length)
    {
      fprintf (stderr, "bytes = %u, content_length = %u\n",
	      tunnel->bytes, tunnel->content_length);
      tunnel_padding (tunnel, tunnel->content_length - tunnel->bytes);
    }

  if (tunnel_is_disconnected (tunnel) &&
      tunnel_out_connect (tunnel) == -1)
    return -1;

  if (tunnel_write_data (tunnel, &request, sizeof request) == -1)
    return -1;

  if (length > 0)
    {
      Length network_length = htons (length);
      if (tunnel_write_data (tunnel,
			     &network_length,
			     sizeof network_length) == -1)
	return -1;

      if (tunnel_write_data (tunnel, data, (size_t)length) == -1)
	return -1;
    }

  if (tunnel->bytes > tunnel->content_length)
    {
      fprintf (stderr, "tunnel_write_request (): tunnel->bytes > tunnel->content_length\n");
    }
		    
  if (tunnel->bytes == tunnel->content_length)
    tunnel_disconnect (tunnel);

  return 0;
}

int
tunnel_connect (Tunnel *tunnel)
{
  char auth_data[1] = { 42 }; /* dummy data, not used */

  if (tunnel->out_fd != -1)
    {
      errno = EINVAL;
      return -1;
    }

  if (tunnel_write_request (tunnel, TUNNEL_OPEN,
			    auth_data, sizeof auth_data) == -1)
    return -1;

  tunnel_in_connect (tunnel);

  return 0;
}

static inline int
tunnel_write_or_padding (Tunnel *tunnel, Request request, void *data,
			 size_t length)
{
  static char padding[65536];
  size_t n, remaining;

  for (remaining = length; remaining > 0; remaining -= n, data += n)
    {
      if (tunnel->bytes + remaining > tunnel->content_length - sizeof_header &&
	  tunnel->content_length - tunnel->bytes > sizeof_header)
	n = tunnel->content_length - sizeof_header - tunnel->bytes;
      else if (remaining > tunnel->content_length - sizeof_header)
	n = tunnel->content_length - sizeof_header;
      else
	n = remaining;

      if (n > 65535)
	n = 65535;

      if (request == TUNNEL_PADDING)
	{
	  if (n + sizeof_header > remaining)
	    n = remaining - sizeof_header;
	  tunnel_write_request (tunnel, request, padding, n);
	  n += sizeof_header;
	}
      else
	tunnel_write_request (tunnel, request, data, n);
    }

  return length;
}

int
tunnel_write (Tunnel *tunnel, void *data, size_t length)
{
  if (tunnel_is_client (tunnel))
    return tunnel_write_or_padding (tunnel, TUNNEL_DATA, data, length);
  else
    return write (tunnel->out_fd, data, length);
}

int
tunnel_padding (Tunnel *tunnel, size_t length)
{
  if (length < sizeof_header + 1)
    {
      int i;

      for (i = 0; i < length; i++)
	tunnel_write_request (tunnel, TUNNEL_PAD1, NULL, 0);
      return 0;
    }

  return tunnel_write_or_padding (tunnel, TUNNEL_PADDING, NULL, length);
}

int
tunnel_close (Tunnel *tunnel)
{
  char buf[10240];
  ssize_t n;

  if (tunnel_is_client (tunnel))
    {
      tunnel_write_request (tunnel, TUNNEL_CLOSE, NULL, 0);
      tunnel_padding (tunnel, tunnel->content_length - tunnel->bytes);
    }

  tunnel_disconnect (tunnel);
  
  while ((n = read (tunnel->in_fd, buf, sizeof buf)) != 0 &&
	 !(n == -1 && errno != EAGAIN))
    ;

  close (tunnel->in_fd);
  tunnel->in_fd = -1;

  tunnel->buf_len = 0;

  return 0;
}

Tunnel *
tunnel_new_client (const char *host, int host_port,
		   const char *proxy, int proxy_port,
		   size_t content_length)
{
  const char *remote;
  int remote_port;
  Tunnel *tunnel;

  tunnel = malloc (sizeof (Tunnel));
  if (tunnel == NULL)
    return NULL;

  tunnel->in_fd = -1;
  tunnel->out_fd = -1;
  tunnel->server_socket = -1;
  tunnel->host = host;
  tunnel->host_port = host_port;
  tunnel->proxy = proxy;
  tunnel->proxy_port = proxy_port;
  tunnel->content_length = content_length;

  if (tunnel->proxy == NULL)
    {
      remote = tunnel->host;
      remote_port = tunnel->host_port;
    }
  else
    {
      remote = tunnel->proxy;
      remote_port = tunnel->proxy_port;
    }

  if (set_address (&tunnel->address, remote, remote_port) == -1)
    {
      free (tunnel);
      return NULL;
    }

  return tunnel;
}

static ssize_t
tunnel_server_read (Tunnel *tunnel, void *data, size_t length)
{
  char buf[65536];
  Request req;
  Length len;
  ssize_t n;

  if (tunnel->buf_len > 0)
    {
      n = min (tunnel->buf_len, length);
      memcpy (data, tunnel->buf_ptr, n);
      tunnel->buf_ptr += n;
      tunnel->buf_len -= n;
      return n;
    }

  if (tunnel->in_fd == -1)
    {
      if (tunnel_accept (tunnel) == -1)
	return -1;

      return tunnel_read (tunnel, data, length);
    }

  n = read (tunnel->in_fd, &req, 1);
  if (n == -1)
    return n;
  else if (n == 0)
    {
      close (tunnel->in_fd);
      tunnel->in_fd = -1;
      errno = EAGAIN;
      return -1;
    }

  switch (req)
    {
    case TUNNEL_OPEN:
      n = read_all (tunnel->in_fd, &len, 2);
      if (n <= 0)
	return -1;
      len = ntohs (len);

      n = read_all (tunnel->in_fd, buf, (size_t)len);
      if (n <= 0)
	return -1;

      /* do something with buf */

      break;

    case TUNNEL_DATA:
      n = read_all (tunnel->in_fd, &len, 2);
      if (n <= 0)
	return -1;
      len = ntohs (len);

      n = read_all (tunnel->in_fd, tunnel->buf, (size_t)len);
      if (n <= 0)
	return -1;

      tunnel->buf_ptr = tunnel->buf;
      tunnel->buf_len = n;

      break;

    case TUNNEL_PADDING:
      n = read_all (tunnel->in_fd, &len, 2);
      if (n <= 0)
	return -1;
      len = ntohs (len);

      n = read_all (tunnel->in_fd, buf, (size_t)len);
      if (n <= 0)
	return -1;

      /* discard data */

      break;

    case TUNNEL_PAD1:
      /* do nothing */
      return tunnel_server_read (tunnel, data, length);

    case TUNNEL_CLOSE:
      return 0;

    default:
      fprintf (stderr, "tunnel_server_read: error in protocol: unknown packet type\n");
      errno = EINVAL;
      return -1;
    }

  errno = EAGAIN;
  return -1;
}

ssize_t
tunnel_read (Tunnel *tunnel, void *data, size_t length)
{
  if (tunnel_is_client (tunnel))
    return read (tunnel->in_fd, data, length);
  else
    return tunnel_server_read (tunnel, data, length);
}

void
tunnel_read_only (Tunnel *tunnel)
{
  tunnel_disconnect (tunnel);
}

void
tunnel_write_only (Tunnel *tunnel)
{
  if (tunnel->in_fd != -1)
    {
      fsync (tunnel->in_fd);
      close (tunnel->in_fd);
    }
  tunnel->in_fd = -1;
}

/*
Return the input file descriptor.

If we're on the server side, and the connection is closed, return
the file descriptor used to accept connections.  This is to be able
to use tunnel_in_fd() in a poll() call even if there is no connection.

FIXME: this isn't very clean.
*/

int
tunnel_in_fd (Tunnel *tunnel)
{
  if (tunnel_is_server (tunnel) &&
      tunnel->in_fd == -1)
    return tunnel->server_socket;
  else
    return tunnel->in_fd;
}

int
tunnel_out_fd (Tunnel *tunnel)
{
  return tunnel->out_fd;
}

/*
If the write connection is up and needs padding to the block length
specified in the second argument, send some padding.
*/

int
tunnel_maybe_pad (Tunnel *tunnel, size_t length)
{
  size_t padding;

  if (tunnel_is_disconnected (tunnel) ||
      tunnel->bytes % length == 0)
    return 0;

  padding = length - tunnel->bytes % length;
  if (padding > tunnel->content_length - tunnel->bytes)
    padding = tunnel->content_length - tunnel->bytes;

  return tunnel_padding (tunnel, padding);
}

int
parse_header (int s)
{
  static const char *end_of_header = "\r\n\r\n";
  ssize_t n;
  int type;
  char c;
  int i;

  n = read_all (s, &c, 1);
  if (n != 1 && errno != EAGAIN)
    return -1;

  if (c == 'P')
    type = TUNNEL_IN;
  else if (c == 'G')
    type = TUNNEL_OUT;
  else
    {
      errno = EINVAL;
      return -1;
    }

  i = 0;
  while (i < 4)
    {
      n = read_all (s, &c, 1);
      if (n != 1 && errno != EAGAIN)
	return n;

      if (c == end_of_header[i])
	i++;
      else
	i = 0;
    }

  return type;
}

int
tunnel_accept (Tunnel *tunnel)
{
printf ("tunnel_accept (%p)\n", tunnel);
  while (tunnel->in_fd == -1 || tunnel->out_fd == -1)
    {
      struct sockaddr_in addr; 
      int type;
      int len;
      int s;

      len = sizeof addr;
      s = accept (tunnel->server_socket, (struct sockaddr *)&addr, &len);
printf ("accept (%d, ...) = %d\n", tunnel->server_socket, s);
      if (s == -1)
	return -1;

      type = parse_header (s);
      if (type == -1)
	{
	  fprintf (stderr, "tunnel_accept: error parsing header: %s\n",
		  strerror (errno));
	  close (s);
	}
      else if (type == TUNNEL_IN)
	{
	  if (tunnel->in_fd == -1)
	    {
	      struct linger l;
#if defined (SO_SNDLOWAT) || defined (SO_RCVLOWAT)
	      int i;
#endif

	      tunnel->in_fd = s;

	      fcntl (tunnel->in_fd,
		     F_SETFL,
		     fcntl (tunnel->in_fd, F_GETFL) | O_NONBLOCK);

#ifdef SO_SNDLOWAT
	      i = 1;
	      setsockopt (tunnel->in_fd,
			  SOL_SOCKET,
			  SO_SNDLOWAT,
			  &i,
			  sizeof i);
#endif

#ifdef SO_RCVLOWAT
	      i = 1;
	      setsockopt (tunnel->in_fd,
			  SOL_SOCKET,
			  SO_RCVLOWAT,
			  &i,
			  sizeof i);
#endif

	      l.l_onoff = 1;
	      l.l_linger = 20 * 100; /* linger for 20 seconds */
	      setsockopt (tunnel->in_fd,
			  SOL_SOCKET,
			  SO_LINGER,
			  &l,
			  sizeof l);
	    }
	  else
	    {
	      fprintf (stderr, "rejected tunnel_in: already got a connection\n");
	      close (s);
	    }
	}
      else if (type == TUNNEL_OUT)
	{
	  if (tunnel->out_fd == -1)
	    {
	      tunnel->out_fd = s;
	    }
	  else
	    {
	      fprintf (stderr, "rejected tunnel_out: already got a connection\n");
	      close (s);
	    }
	}
      else
	{
	  fprintf (stderr, "tunnel_accept: unknown header type\n");
	  fprintf (stderr, "tunnel_accept: closing connection\n");
	  close (s);
	}
    }

  return 0;
}

Tunnel *
tunnel_new_server (int port)
{
  Tunnel *tunnel;

  tunnel = malloc (sizeof (Tunnel));
  if (tunnel == NULL)
    return NULL;

  tunnel->in_fd = -1;
  tunnel->out_fd = -1;
  tunnel->server_socket = -1;
  tunnel->host = NULL;
  tunnel->host_port = port;
  tunnel->buf_ptr = tunnel->buf;
  tunnel->buf_len = 0;

  tunnel->server_socket = server_socket (tunnel->host_port);
  if (tunnel->server_socket == -1)
    {
      tunnel_destroy (tunnel);
      return NULL;
    }

  return tunnel;
}

void
tunnel_destroy (Tunnel *tunnel)
{
  if (tunnel_is_connected (tunnel) || tunnel->in_fd != -1)
    tunnel_close (tunnel);

  if (tunnel->server_socket != -1)
    close (tunnel->server_socket);

  free (tunnel);
}
